#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****


_ATTR_BOLD       = 1
_ATTR_UNDERLINED = 4
_ATTR_BLINK      = 5
_ATTR_INVERSE    = 7
_ATTR_INVISIBLE  = 8
_ATTR_FG         = 14 
_ATTR_BG         = 23 

class Attribute():

    __value = (0x100 << _ATTR_FG | 0x100 << _ATTR_BG, 0x42)

    def __init__(self, value = (0x100 << _ATTR_FG | 0x100 << _ATTR_BG, 0x42)):
        self.__value = value

    def draw(self, s):
        params = [0]
        value, charset = self.__value 
        for i in xrange(0, 8):
            if value & (2 << i):
                params.append(i) 

        fg = value >> _ATTR_FG & 0x1ff
        if fg == 0:
            pass
        elif fg < 8:
            params.append(30 + fg)
        elif fg == 0x100:
            params.append(39)
        else:
            params += [38, 5, fg]

        bg = value >> _ATTR_BG & 0x1ff
        if fg == 0:
            pass
        elif bg < 8:
            params.append(40 + bg)
        elif bg == 0x100:
            params.append(49)
        else:
            params += [48, 5, bg]
        s.write(u'\x1b(%c\x1b[%sm' % (charset, ';'.join([str(p) for p in params])))

    def clear(self):
        self.__value = (0x100 << _ATTR_FG | 0x100 << _ATTR_BG, 0x42)

    def get(self):
        return self.__value

    def set_charset(self, charset):
        value, old = self.__value
        self.__value = (value, charset)

    def set_sgr(self, pm):
        i = 0
        if len(pm) == 0:
            pm.append(0)
        value, charset = self.__value
        while i < len(pm):
            n = pm[i]
            if n == 0:
                value = 0x100 << _ATTR_FG | 0x100 << _ATTR_BG
            elif n == 1:
                value |= 2 << _ATTR_BOLD 
            elif n == 4:
                value |= 2 << _ATTR_UNDERLINED 
            elif n == 5:
                value |= 2 << _ATTR_BLINK 
            elif n == 7:
                value |= 2 << _ATTR_INVERSE 
            elif n == 8:
                value |= 2 << _ATTR_INVISIBLE 
            elif n == 21:
                value &= ~(2 << _ATTR_BOLD)
            elif n == 22:
                value &= ~(2 << _ATTR_BOLD | 2 << _ATTR_UNDERLINED)
            elif n == 24:
                value &= ~(2 << _ATTR_UNDERLINED)
            elif n == 25:
                value &= ~(2 << _ATTR_BLINK)
            elif n == 27:
                value &= ~(2 << _ATTR_INVERSE)
            elif n == 28:
                value &= ~(2 << _ATTR_VISIBLE)
            elif 30 <= n and n < 38:
                value = value & ~(0x1ff << _ATTR_FG) | (n - 30) << _ATTR_FG

            elif n == 38:
                i += 1
                n1 = pm[i]
                i += 1
                n2 = pm[i]
                if n1 == 5:
                    value = value & ~(0x1ff << _ATTR_FG) | (n2 << _ATTR_FG)
            elif n == 39:
                value = value & ~(0x1ff << _ATTR_FG) | (0x100 << _ATTR_FG)
            elif 40 <= n and n < 48:
                value = value & ~(0x1ff << _ATTR_BG) | (n - 40) << _ATTR_BG

            elif n == 48:
                i += 1
                n1 = pm[i]
                i += 1
                n2 = pm[i]
                if n1 == 5:
                    value = value & ~(0x1ff << _ATTR_BG) | (n2 << _ATTR_BG)
            elif n == 49:
                value = value & ~(0x1ff << _ATTR_BG) | (0x100 << _ATTR_BG)
            elif 90 <= n and n < 98:
                value = value & ~(0x1ff << _ATTR_FG) | (n - 90 + 8) << _ATTR_FG
            elif 100 <= n and n < 108:
                value = value & ~(0x1ff << _ATTR_BG) | (n - 100 + 8) << _ATTR_BG
            else:
               pass
               #logger.writeLine("SGR %d is ignored." % n)
            i += 1
        self.__value = (value, charset)

    def equeals(self, other):
        return self.__value == other.__value

    def __str__(self):
        import StringIO
        s = StringIO.StringIO()
        self.draw(s)
        return s.getvalue().replace("\x1b", "<ESC>")

def test():
    attr = Attribute() 
    print attr
    attr.set_sgr([0])
    print attr.equeals(Attribute())
    print attr
    attr.set_charset("A")
    print attr
    attr.set_sgr([0, 5, 6])
    print attr
    attr.set_sgr([7, 8])
    print attr
    attr.set_sgr([17, 18])
    print attr
    attr.set_sgr([38, 5, 200, 48, 5, 100])
    print attr

if __name__ == "__main__":
    print "attribute test."
    test()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

def _printver():
        import __init__
        print '''
canossa %s
Copyright (C) 2012 Hayaki Saito <user@zuse.jp>. 

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see http://www.gnu.org/licenses/.
        ''' % __init__.__version__
        return

def create(row, col, y, x, is_cjk, visibility=False):
    import output, screen

    screen = screen.Screen(row, col, y, x, is_cjk)
    return output.OutputHandler(screen, visibility=visibility)

def main():
    import sys, os, optparse, select
    # parse options and arguments
    usage = 'usage: %prog [options] command'
    parser = optparse.OptionParser(usage=usage)

    parser.add_option('-v', '--visible', dest='visibility',
                      action="store_true", default=False,
                      help='bring up the front of terminal (visible mode)')

    parser.add_option('--version', dest='version',
                      action="store_true", default=False,
                      help='show version')

    (options, args) = parser.parse_args()

    # print version
    if options.version:
        _printver()
        return

    # retrive starting command
    if len(args) > 0:
        command = args[0]
    elif not os.getenv('SHELL') is None:
        command = os.getenv('SHELL')
    else:
        command = '/bin/sh'

    # retrive TERM setting
    if not os.getenv('TERM') is None:
        term = os.getenv('TERM')
    else:
        term = 'xterm'

    # retrive LANG setting
    if not os.getenv('LANG') is None:
        lang = os.getenv('LANG')
    else:
        import locale
        lang = '%s.%s' % locale.getdefaultlocale()

    # retrive terminal encoding setting
    import locale
    language, encoding = locale.getdefaultlocale()
    termenc = encoding

    import tff, output

    # create pty
    tty = tff.DefaultPTY(term, lang, command, sys.stdin)

    # fit to screen and get size
    tty.fitsize()

    # make screen
    if options.visibility:
        outputhandler = output.OutputHandler(visibility=True)
    else:
        canossahandler = output.OoutputHandler(visibility=False)
        outputhandler = tff.FilterMultiplexer(canossahandler, tff.DefaultHandler())

    # create TFF session
    session = tff.Session(tty)

    # start session
    session.start(termenc=termenc,
                  stdin=sys.stdin,
                  stdout=sys.stdout,
                  outputhandler=outputhandler,
                  buffering=False)
        

''' main '''
if __name__ == '__main__':    
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

class Cell():

    __value = None
    attr = (0, 0x42)

    def __init__(self):
        self.__value = u' '

    def write(self, value, attr):
        self.__value = unichr(value)
        self.attr = attr.get()

    def pad(self):
        self.__value = None 

    def combine(self, value):
        self.__value += unichr(value)

    def get(self):
        return self.__value

    def clear(self, attr):
        self.__value = u' '
        self.attr = attr.get()

def test():
    from attribute import Attribute
    cell = Cell() 
    attr = Attribute()
    print cell.get()
    cell.write(0x34, attr)
    print cell.get()
    cell.clear(attr)
    print cell.get()
    cell.write(0x3042, attr)
    print cell.get()
    cell.pad()
    print cell.get()
    cell.write(0x09a4, attr)
    print cell.get()
    cell.combine(0x20DE)
    print cell.get()
    cell.combine(0x20DD)
    print cell.get()
    cell.combine(0x0308)
    print cell.get()

if __name__ == "__main__":
    print "line test."
    test()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****

from attribute import Attribute
from cell import Cell

_LINE_TYPE_DHLT = 3
_LINE_TYPE_DHLB = 4
_LINE_TYPE_SWL  = 5
_LINE_TYPE_DWL  = 6


class SupportsDoubleSizedTrait():

    _type = _LINE_TYPE_SWL

    def set_swl(self):
        self._type = _LINE_TYPE_SWL
        self.dirty = True

    def set_dwl(self):
        self._type = _LINE_TYPE_DWL
        self.dirty = True

    def set_dhlt(self):
        self._type = _LINE_TYPE_DHLT
        self.dirty = True

    def set_dhlb(self):
        self._type = _LINE_TYPE_DHLB
        self.dirty = True

    def type(self):
        return self._type

class SupportsWideTrait():

    def pad(self, pos):
        self.cells[pos].pad()

class SupportsCombiningTrait():

    def combine(self, value, pos):
        self.cells[max(0, pos - 1)].combine(value)

class Line(SupportsDoubleSizedTrait,
           SupportsWideTrait,
           SupportsCombiningTrait):

    def __init__(self, width):
        self.cells = [Cell() for cell in xrange(0, width)]
        self.dirty = True

    def resize(self, col):
        width = len(self.cells)
        if col < width:
            self.cells = self.cells[:col]
        elif col > width:
            self.cells += [Cell() for cell in xrange(0, col - width)]
        self.dirty = True

    def clear(self, attr):
        self.dirty = True
        self._type = _LINE_TYPE_SWL
        for cell in self.cells:
            cell.clear(attr)

    def write(self, value, pos, attr):
        self.cells[pos].write(value, attr)

    def draw(self, s, left, right):
        self.dirty = False
        attr = None 
        cells = self.cells
        c = None
        s.write("\x1b#%d" % self._type)
        if left > 0:
            c = cells[left - 1].get()
            if c is None:
                s.write(u'\x08') # BS
                left -= 1
        for cell in cells[left:right]:
            c = cell.get()
            if not c is None:
                if attr != cell.attr:
                    attr = cell.attr
                    Attribute(attr).draw(s)
                s.write(c)
        if c is None:
            for cell in cells[right:]:
                c = cell.get()
                if not c is None:
                    if attr != cell.attr:
                        attr = cell.attr
                        Attribute(attr).draw(s)
                    s.write(c)
                    break

    def __str__(self):
        import StringIO, codecs
        import locale
        language, encoding = locale.getdefaultlocale()
        s = codecs.getwriter(encoding)(StringIO.StringIO())
        self.draw(s, 0, len(self.cells)) 
        return s.getvalue().replace("\x1b", "<ESC>").replace("\x00", "<NUL>")

def test():
    line = Line(10) 
    attr = Attribute()
    print line
    line.clear(attr)
    print line
    line.write(0x40, 0, attr)
    line.write(0x50, 0, attr)
    print line
    line.write(0x40, 1, attr)
    print line
    line.write(0x3042, 2, attr)
    print line
    line.write(0x30, 5, attr)
    print line

if __name__ == "__main__":
    print "line test."
    test()




# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from canossa import __version__, __license__, __author__

setup(name                  = 'canossa',
      version               = __version__,
      description           = 'Provides basic, transparent, off-screen(invisible) terminal emulation service, for terminal apps.',
      long_description      = open("README.rst").read(),
      py_modules            = ['canossa'],
      eager_resources       = [],
      classifiers           = ['Development Status :: 4 - Beta',
                               'Topic :: Terminals',
                               'Environment :: Console',
                               'Intended Audience :: Developers',
                               'License :: OSI Approved :: GNU General Public License (GPL)',
                               'Programming Language :: Python'
                               ],
      keywords              = 'terminal',
      author                = __author__,
      author_email          = 'user@zuse.jp',
      url                   = 'https://github.com/saitoha/canossa',
      license               = __license__,
      packages              = find_packages(exclude=[]),
      zip_safe              = True,
      include_package_data  = False,
      install_requires      = ['tff ==0.0.9'],
      entry_points          = """
                              [console_scripts]
                              canossa = canossa:main
                              """
      )




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ***** BEGIN LICENSE BLOCK *****
# Copyright (C) 2012  Hayaki Saito <user@zuse.jp>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ***** END LICENSE BLOCK *****


import wcwidth
try:
    from CStringIO import StringIO
except:
    from StringIO import StringIO
import sys
#import logger

import sys, os, termios, select
from attribute import Attribute

#
# CSI ... ; ... R
#
def _get_pos():
    stdin = sys.stdin 
    stdout = sys.stdout
    stdin_fileno = stdin.fileno()
    vdisable = os.fpathconf(stdin_fileno, 'PC_VDISABLE')
    backup = termios.tcgetattr(stdin_fileno)
    new = termios.tcgetattr(stdin_fileno)
    new[3] &= ~(termios.ECHO | termios.ICANON)
    termios.tcsetattr(stdin_fileno, termios.TCSANOW, new)
    try:
        stdout.write("\x1b[6n")
        stdout.flush()
    except:
        pass
    def get_report():
        
        rfd, wfd, xfd = select.select([stdin_fileno], [], [], 0.5)
        if rfd:
            data = os.read(stdin_fileno, 1024)
            assert data[:2] == '\x1b['
            assert data[-1] == 'R'
            y, x = [int(n) - 1 for n in  data[2:-1].split(';')]
            return y, x
    try:
        return get_report()
    except:
        import time
        time.sleep(0.1)
        return get_report()
    finally:
        termios.tcsetattr(stdin_fileno, termios.TCSANOW, backup)

from cursor import Cursor
from line import Line

class SupportsDoubleSizedTrait():

    def decdhlt(self):
        line = self.lines[self.cursor.row] 
        line.set_dhlt()

    def decdhlb(self):
        line = self.lines[self.cursor.row] 
        line.set_dhlb()

    def decswl(self):
        line = self.lines[self.cursor.row] 
        line.set_swl()

    def decdwl(self):
        line = self.lines[self.cursor.row] 
        line.set_dwl()

class SupportsWideCharacterTrait():

    _wcwidth = None

    def _setup_wcwidth(self, is_cjk=False):
        if is_cjk:
            self._wcwidth = wcwidth.mk_wcwidth
        else:
            self._wcwidth = wcwidth.mk_wcwidth_cjk

class SuuportsCursorPersistentTrait():

    def save_pos(self):
        self._saved_pos = (self.cursor.row, self.cursor.col)

    def restore_pos(self):
        if self._saved_pos:
            self.cursor.row, self.cursor.col = self._saved_pos

class SuuportsISO2022DesignationTrait():

    __g = None 
    __gl = None

    def _setup_charset(self):
        self.__g = [0x42, 0x30, 0x42, 0x42]
        self.__gl = self.__g[0]

    def set_g0(self, c):
        self.__g[0] = c
        self.cursor.attr.set_charset(c)

    def set_g1(self, c):
        self.__g[1] = c
        self.cursor.attr.set_charset(c)

    def so(self):
        self.__gl = self.__g[1]
        self.cursor.attr.set_charset(self.__gl)

    def si(self):
        self.__gl = self.__g[0]
        self.cursor.attr.set_charset(self.__gl)


class SuuportsAlternateScreenTrait():

    def _setup_altbuf(self):
        self._altbuf = [Line(self.width) for line in xrange(0, self.height)]
        self._mainbuf = self.lines

    def switch_mainbuf(self):
        for line in self.lines:
            line.clear(self.cursor.attr)
        self.lines = self._mainbuf
        for line in self.lines:
            line.resize(self.width)

    def switch_altbuf(self):
        self.lines = self._altbuf
        for line in self.lines:
            line.clear(self.cursor.attr)
        lines = self.lines
        if len(lines) > self.height:
            while len(lines) != self.height:
                lines.pop()
            for line in lines:
                line.resize(self.width)
        elif len(lines) < self.height:
            for line in lines:
                line.resize(self.width)
            while len(lines) != self.height:
                lines.insert(0, Line(self.width))
        else:
            for line in lines:
                line.resize(self.width)
        assert len(lines) == self.height
        for line in lines:
            assert self.width == len(line.cells)

class SupportsAnsiModeTrait():
    pass

class SupportsExtendedModeTrait():

    dectcem = True
    decawm = True
    decom = False

    def decset(self, params):
        for param in params:
            if param == 1:
                pass # DECCKM
            elif param == 3:
                self.resize(self.width, 132)
                self.ris()
            elif param == 4:
                pass
            elif param == 6:
                self.decom = True
            elif param == 7:
                self.decawm = True
            elif param == 12:
                pass # cursor blink
            elif param == 25:
                self.dectcem = True
            elif param == 1047:
                self.switch_altbuf()
                return True
            elif param == 1048:
                self.save_pos()
                return True
            elif param == 1049:
                self.save_pos()
                self.switch_altbuf()
                return True
            else:
                # 1000 normal mouse tracking 
                # 1001 highlight mouse tracking 
                # 1002 button mouse tracking 
                # 1003 all mouse tracking 
                # 1005 UTF-8 mouse encoding 
                # 1006 SGR mouse encoding 
                # 1015 URXVT mouse encoding 
                # 2004 bracketed paste mode
                pass
                #import tff
                #raise tff.NotHandledException("DECSET %d" % param)

    def decrst(self, params):
        for param in params:
            if param == 1:
                pass # DECCKM
            elif param == 3:
                self.resize(self.width, 80)
                self.ris()
            elif param == 4:
                pass
            elif param == 6:
                self.decom = False
            elif param == 7:
                self.decawm = False
            elif param == 12:
                pass # cursor blink
            elif param == 25:
                self.dectcem = False
            elif param == 1000:
                pass
            elif param == 1047:
                self.switch_mainbuf()
                return True
            elif param == 1048:
                self.restore_pos()
                return True
            elif param == 1049:
                self.switch_mainbuf()
                self.restore_pos()
                return True
            elif param == 2004:
                pass # bracketed paste mode

    def xt_save(self, params):
        pass

    def xt_rest(self, params):
        pass


class SupportsTabStopTrait():

    _tabstop = None

    def _setup_tab(self):
        self._tabstop = [n for n in xrange(0, self.width + 1, 8)] 

    def hts(self):
        if self.cursor.col in self._tabstop:
            pass
        elif len(self._tabstop) > 0 and self._tabstop[0] > self.cursor.col:
            self._tabstop.insert(0, self.cursor.col)
        else:
            for i, stop in enumerate(self._tabstop[0:]):
                if self.cursor.col < stop:
                    self._tabstop.insert(i, self.cursor.col)
                    break
            else:
                self._tabstop.append(self.cursor.col)

    def tbc(self, ps):
        if ps == 0:
            if self.cursor.col in self._tabstop:
                self._tabstop.remove(self.cursor.col)
        elif ps == 3:
            self._tabstop = []

    def ht(self):
        line = self.lines[self.cursor.row] 
        col = self.cursor.col
        if col < self.width:
            col += 1
            for stop in self._tabstop:
                if col <= stop:
                    self.cursor.col = stop 
                    break
            else:
                self.cursor.col = self.width - 1 
        else: 
            self.cursor.col = 0
        self.cursor.dirty = True

from interface import *

class ICanossaScreenImpl(ICanossaScreen):

    def drawrect(self, s, col, row, width, height):
        height = min(height, self.height)
        width =  min(width, self.width)
        for i in xrange(row, row + height):
            s.write("\x1b[%d;%dH" % (i + 1, col + 1))
            self.lines[i].draw(s, col, col + width)
        self.cursor.attr.draw(s)

    def draw(self, context):
        cursor = self.cursor
        s = StringIO()
        self.drawrect(s, 0, 0, self.width, self.height)
        cursor.draw(s)
        context.writestring(s.getvalue())

    def resize(self, row, col):
        lines = self.lines
        height = len(lines)
        if row < height:
            while row != len(lines):
                lines.pop()
            for line in lines:
                line.resize(col)
        elif row > height:
            for line in lines:
                line.resize(col)
            while row != len(lines):
                lines.insert(0, Line(col))
        else:
            for line in lines:
                line.resize(col)
        assert row == len(lines)
        for line in lines:
            assert col == len(line.cells)
        if self.scroll_top == 0 and self.scroll_bottom == self.height:
            self.scroll_top = 0
            self.scroll_bottom = row 
        self.height = row
        self.width = col
        try:
        #    sys.stdout.write("\x1b]2;%d-%d\x1b\\" % (row, col))
            self.cursor.row, self.cursor.col = _get_pos()
        except:
            pass
        if self.cursor.row >= self.height:
            self.cursor.row = self.height - 1
        if self.cursor.col >= self.width:
            self.cursor.col = self.width - 1
        self._setup_tab()

    def write(self, c):
        row, col = self.cursor.row, self.cursor.col
        line = self.lines[row] 

        if col >= self.width:
            self._wrap()
            row, col = self.cursor.row, self.cursor.col
            line = self.lines[row] 

        line.dirty = True

        width = self._wcwidth(c)

        if width == 1: # normal (narrow) character
            if self.cursor.col >= self.width:
                self.cursor.col = self.width - 1
                col = self.cursor.col
            line.write(c, col, self.cursor.attr)
            self.cursor.dirty = True
            self.cursor.col += 1
            #if col > len(line.cells) - 1:
            #    self.cursor.col = len(line.cells) - 1
        elif width == 2: # wide character
            if self.cursor.col >= self.width - 1:
                self.cursor.col = self.width - 2
                col = self.cursor.col
            line.pad(col)
            line.write(c, col + 1, self.cursor.attr)
            self.cursor.dirty = True
            self.cursor.col += 2
            #if col > len(line.cells) - 1:
            #    self.cursor.col = len(line.cells) - 1
        elif width == 0: # combining character
            line.combine(c, col)

class MockScreen():

    width = 80
    height = 24

    def __init__(self, row=24, col=80):
        self.height = row
        self.width = col
        self._setup_lines()

    def _setup_lines(self):
        self.lines = [Line(self.width) for line in xrange(0, self.height)]

class MockScreenWithCursor(MockScreen):

    width = 80
    height = 24

    def __init__(self, row=24, col=80, y=0, x=0):
        self.height = row
        self.width = col
        self.cursor = Cursor(y, x)
        self._setup_lines()


class Screen(ICanossaScreenImpl,
             MockScreenWithCursor,
             SupportsAnsiModeTrait,
             SupportsExtendedModeTrait,
             SupportsTabStopTrait,
             SupportsDoubleSizedTrait,
             SupportsWideCharacterTrait,
             SuuportsCursorPersistentTrait,
             SuuportsAlternateScreenTrait,
             SuuportsISO2022DesignationTrait):

    _saved_pos = None

    def __init__(self, row=24, col=80, y=0, x=0, is_cjk=False):
        self.height = row
        self.width = col
        self.cursor = Cursor(y, x)
        self.scroll_top = 0 
        self.scroll_bottom = self.height 

        self._setup_lines()
        self._setup_wcwidth()
        self._setup_altbuf()
        self._setup_tab()
        self._setup_charset()

    def _wrap(self):
        self.cursor.col = 0 
        if self.decawm:
            self.lf()

    def bs(self):
        if self.cursor.col >= self.width:
            self.cursor.col = self.width - 1

        if self.cursor.col <= 0:
            pass
        else:
            self.cursor.col -= 1
        self.cursor.dirty = True

    def lf(self):
        if self.cursor.col >= self.width:
            self._wrap() 
        self.cursor.row += 1
        if self.cursor.row >= self.scroll_bottom:
            for line in self.lines:
                line.dirty = True
            self.lines.insert(self.scroll_bottom, Line(self.width))
            self.lines.pop(self.scroll_top)
            self.cursor.row = self.scroll_bottom - 1 
        self.cursor.dirty = True

    def ind(self):
        self.lf()

    def nel(self):
        self.cr()
        self.lf()

    def ri(self):
        if self.cursor.row <= self.scroll_top:
            for line in self.lines:
                line.dirty = True
            self.lines.insert(self.scroll_top, Line(self.width))
            self.lines.pop(self.scroll_bottom)
            self.cursor.row = self.scroll_top 
        else:
            self.cursor.row -= 1
        self.cursor.dirty = True

    def cr(self):
        self.cursor.col = 0
        self.cursor.dirty = True
     
    def decstbm(self, top, bottom):
        self.scroll_top = max(0, top)
        self.scroll_bottom = min(bottom + 1, self.height)
        self.cursor.row = self.scroll_top
        self.cursor.col = 0

    def dch(self, n):
        self.cursor.dirty = True

    def vpa(self, row):
        if row >= self.scroll_bottom:
            self.cursor.row = self.scroll_bottom - 1
        elif row < self.scroll_top:
            self.cursor.row = self.scroll_top
        else:
            self.cursor.row = row
        self.cursor.dirty = True

    def hvp(self, row, col):
        if row >= self.scroll_bottom:
            self.cursor.row = self.scroll_bottom - 1
        elif row < self.scroll_top:
            self.cursor.row = self.scroll_top
        else:
            self.cursor.row = row
        self.cursor.col = col
        self.cursor.dirty = True

    def cup(self, row, col):
        if self.decom:
            row += self.scroll_top
            top = self.scroll_top
            bottom = self.scroll_bottom
        else:
            top = 0
            bottom = self.height
        if row >= bottom:
            self.cursor.row = bottom - 1
        elif row < top:
            self.cursor.row = top
        else:
            self.cursor.row = row
        self.cursor.col = col
        self.cursor.dirty = True

    def ed(self, ps):
        if self.cursor.row >= self.height:
            self.cursor.row = self.height - 1
        if self.cursor.col >= self.width:
            self.cursor.col = self.width - 1
        if ps == 0:
            line = self.lines[self.cursor.row] 
            line.dirty = True
            for cell in line.cells[self.cursor.col:]:
                cell.clear(self.cursor.attr)
            if self.cursor.row < self.height:
                for line in self.lines[self.cursor.row + 1:]:
                    line.clear(self.cursor.attr)

        elif ps == 1:
            line = self.lines[self.cursor.row] 
            line.dirty = True
            for cell in line.cells[:self.cursor.col]:
                cell.clear(self.cursor.attr)
            if self.cursor.row > 0:
                for line in self.lines[:self.cursor.row]:
                    line.clear(self.cursor.attr)

        elif ps == 2:
            for line in self.lines:
                line.clear(self.cursor.attr)

        else:
            raise
  

    def decaln(self):
        for line in self.lines:
            line.dirty = True
            for cell in line.cells:
                cell.write(0x45, self.cursor.attr) # E
        self.scroll_top = 0 
        self.scroll_bottom = self.height 

    def ris(self):
        for line in self.lines:
            line.dirty = True
            for cell in line.cells:
                cell.write(0x20, self.cursor.attr) # SP 
        self.dectcem = True
        self.cursor.clear() 
        self._setup_tab()

    def sgr(self, pm):
        self.cursor.attr.set_sgr(pm)

    def ich(self, ps):
        ''' insert blank character(s) '''    
        row = self.cursor.row
        col = self.cursor.col
        if row >= self.scroll_bottom:
            self.cursor.row = self.scroll_bottom - 1
        elif row < self.scroll_top:
            self.cursor.row = self.scroll_top
        if col >= self.width:
            self.cursor.col = self.width - 1
        cells = self.lines[row].cells

        if col > 0 and cells[col - 1].get() == '\x00':
            col -= 1

        for i in xrange(0, ps):
            cell = cells.pop()
            cell.clear(self.cursor.attr)
            cells.insert(col, cell)

    def cuu(self, ps):
        ''' cursor up '''
        if self.cursor.row >= self.scroll_top + ps:
            self.cursor.row -= ps 
        else:
            self.cursor.row = self.scroll_top 

    def cud(self, ps):
        if self.cursor.row < self.scroll_bottom - ps:
            self.cursor.row += ps 
        else:
            self.cursor.row = self.scroll_bottom - 1 

    def cuf(self, ps):
        if self.cursor.col < self.width - ps:
            self.cursor.col += ps 
        else:
            self.cursor.col = self.width - 1 

    def cub(self, ps):
        if self.cursor.col >= ps:
            self.cursor.col -= ps 
        else:
            self.cursor.col = 0 

    def dl(self, ps):
        row = self.cursor.row
        lines = self.lines
        bottom = self.scroll_bottom
        for line in lines[row + ps:bottom]:
            line.dirty = True
        for x in xrange(0, ps):
            lines.insert(bottom, Line(self.width))
            lines.pop(row)

    def il(self, ps):
        row = self.cursor.row
        lines = self.lines
        bottom = self.scroll_bottom
        for line in lines[row:bottom - ps]:
            line.dirty = True
        for x in xrange(0, ps):
            lines.pop(bottom - 1)
            lines.insert(row, Line(self.width))

    def el(self, ps):
        line = self.lines[self.cursor.row] 
        if ps == 0:
            cells = line.cells[self.cursor.col:]
        elif ps == 1:
            cells = line.cells[:self.cursor.col]
        elif ps == 2:
            cells = line.cells
        else:
            raise
        line.dirty = True
        for cell in cells:
            cell.clear(self.cursor.attr) 

